from .useragents import useragents, newagent
from .session import Session, AlreadyVisitedError
from .tree import bs4, reliq, lxml, html5_parser, lexbor, modest
from .args import args_section, args_session
from .logger import simple_logger



from typing import Optional, Tuple, Any, Callable
import copy
import time
import random
from threading import Lock

from .useragents import newagent
from .logger import create_logger


class AlreadyVisitedError(Exception):
    pass


class RedirectionError(Exception):
    pass


def smarttrim(src):
    # turns " \v i am a   \a \n \n \a test  \a " to "i am a test"

    return " ".join(
        src.translate(str.maketrans("\t\n\r\a\v\f\v", "       ", "")).split()
    )


def newbrowser(browser: Optional[str | Callable]) -> dict:
    if browser is None:
        return {}

    if isinstance(browser, Callable):
        return browser()

    import browser_cookie3

    return getattr(browser_cookie3, browser)()


def _ua(user_agent):
    if isinstance(user_agent, str):
        return [user_agent]
    return user_agent


def Session(
    lib,
    session,
    tree,
    alreadyvisitederror=None,
    requesterror=None,
    redirectionerror=None,
    **kwargs,
):
    orig_tree = tree

    class ret_obj(session):
        def get_settings(
            self, settings: dict, dest: dict = {}, remove: bool = True
        ) -> dict:
            other = dest != self._settings

            user_agent = settings.get("user_agent", -1)
            browser = settings.get("browser", -1)
            logger = settings.get("logger", -1)

            for i in self._settings.keys():
                val = settings.get(i, self._settings[i])

                if i == "headers" or i == "cookies" or i == "proxies":
                    if dest.get(i) is None:
                        dest[i] = {}
                    dest[i].update(val)
                else:
                    dest[i] = val

                if remove:
                    settings.pop(i, None)

            if other and user_agent != -1:
                dest["headers"].update(
                    {"User-Agent": newagent(*_ua(dest["user_agent"]))}
                )
            if other and browser != -1:
                dest["cookies"].update(newbrowser(dest["browser"]))
            if logger != -1:
                dest["_logger"] = create_logger(logger)
            return dest

        def _settings_update(self):
            self.proxies.update(self["proxies"])
            self["proxies"] = {}
            self.headers.update(self["headers"])
            self["headers"] = {}
            self.cookies.update(self["cookies"])
            self["cookies"] = {}

        def set_settings(self, settings: dict, remove=True):
            self.get_settings(settings, dest=self._settings, remove=remove)
            self._settings_update()

        def __getitem__(self, key):
            return self._settings[key]

        def new_user_agent(self):
            self.headers.update({"User-Agent": newagent(*_ua(self["user_agent"]))})

        def new_browser(self):
            self.cookies.update(newbrowser(self["browser"]))

        def __setitem__(self, item, value):
            x = self._settings[item]
            self._settings[item] = value
            if x != value:
                if item == "user_agent":
                    self.new_user_agent()
                if item == "browser":
                    self.new_browser()
                if item == "logger":
                    self._settings["_logger"] = create_logger(value)

        def __init__(self, **settings):
            self._settings = {
                "proxies": {},
                "headers": {},
                "cookies": {},
                "timeout": 30,
                "verify": True,
                "allow_redirects": False,
                "redirects": False,
                "retries": 2,
                "retry_wait": 5,
                "force_retry": False,
                "wait": 0,
                "wait_random": 0,
                "trim": False,
                "user_agent": [
                    ("desktop", "windows", ("firefox", "chrome")),
                ],
                "raise": True,
                "browser": None,
                "visited": False,
                "logger": None,
                "_logger": None,  # compiled .logger
            }
            self.get_settings(settings, dest=self._settings)
            self.visited = set()
            self._lock = Lock()

            super().__init__(**settings)

            self._settings_update()

            self.new_user_agent()
            self.new_browser()

        def new(self, independent=False, **settings):
            r = type(self)(**settings)
            r.headers = copy.copy(self.headers)
            r.cookies = copy.copy(self.cookies)
            r.proxies = copy.copy(self.proxies)

            if independent:
                r.visited = copy.copy(self.visited)
                if r["_logger"] == self["_logger"]:
                    r["_logger"] = None
                    r["logger"] = None
            else:
                r._lock = self._lock
            return r

        def r_req_try(
            self,
            url: str,
            method: str,
            settings: dict,
            kwargs: dict,
            retry: bool = False,
        ):
            if not retry:
                if settings["wait"] != 0:
                    time.sleep(settings["wait"])
                if settings["wait_random"] != 0:
                    time.sleep(random.randint(0, settings["wait_random"] + 1) / 1000)

            visited = settings["visited"]
            if (not retry and visited is not None) or settings["_logger"] is not None:
                with self._lock:
                    if not retry and visited is not None:
                        self.visited.add(url)
                    if settings["_logger"] is not None:
                        settings["_logger"](method, url, retry)

            return getattr(super(), method)(
                url,
                verify=settings["verify"],
                allow_redirects=settings["allow_redirects"],
                timeout=settings["timeout"],
                proxies=settings["proxies"],
                headers=settings["headers"],
                cookies=settings["cookies"],
                **kwargs,
            )

        def r_req(self, url: str, kwargs: dict, method: str = "get"):
            settings = self.get_settings(kwargs)

            tries = settings["retries"]
            retry_wait = settings["retry_wait"]
            if settings["visited"]:
                with self._lock:
                    if url in self.visited:
                        if alreadyvisitederror is not None:
                            raise alreadyvisitederror(url)
                        else:
                            raise AlreadyVisitedError(url)

            instant_end_code = [400, 401, 402, 403, 404, 410, 412, 414, 421, 505]

            i = 0
            while True:
                resp = None
                try:
                    resp = self.r_req_try(url, method, settings, kwargs, retry=(i != 0))
                except (
                    lib.ConnectTimeout,
                    lib.ConnectionError,
                    lib.ReadTimeout,
                    lib.exceptions.ChunkedEncodingError,
                ) as e:
                    if i > tries:
                        if requesterror:
                            raise requesterror()
                        else:
                            raise e

                    ok = False
                else:
                    ok = resp.ok

                if ok:
                    if (
                        not settings["redirects"]
                        and resp.status_code >= 300
                        and resp.status_code < 400
                    ):
                        if redirectionerror:
                            raise redirectionerror()
                        else:
                            raise RedirectionError()

                    return resp

                if (
                    resp is not None
                    and not settings["force_retry"]
                    and resp.status_code in instant_end_code
                ) or i > tries:
                    if settings["raise"]:
                        try:
                            resp.raise_for_status()
                        except Exception as e:
                            if requesterror:
                                raise requesterror()
                            else:
                                raise e

                    return resp

                i += 1
                if retry_wait != 0:
                    time.sleep(retry_wait)

        def get(self, url: str, **settings):
            return self.r_req(url, settings, method="get")

        def post(self, url: str, **settings):
            return self.r_req(url, settings, method="post")

        def head(self, url: str, **settings):
            return self.r_req(url, settings, method="head")

        def put(self, url: str, **settings):
            return self.r_req(url, settings, method="put")

        def delete(self, url: str, **settings):
            return self.r_req(url, settings, method="delete")

        def options(self, url: str, **settings):
            return self.r_req(url, settings, method="options")

        def patch(self, url: str, **settings):
            return self.r_req(url, settings, method="patch")

        def get_html(
            self, url: str, response: bool = False, tree: Callable = None, **settings
        ) -> Any | Tuple[Any, Any]:
            settings = self.get_settings(settings)
            resp = self.get(url, **settings)

            text = resp.text
            if settings["trim"]:
                text = smarttrim(text)

            if tree is None:
                tree = orig_tree
            r = tree(text, url)

            return (r, resp) if response else r

        def get_json(self, url: str, **settings) -> dict:
            resp = self.get(url, **settings)
            return resp.json()

        def post_json(self, url: str, **settings) -> dict:
            resp = self.post(url, **settings)
            return resp.json()

        def delete_json(self, url: str, **settings) -> dict:
            resp = self.delete(url, **settings)
            return resp.json()

        def put_json(self, url: str, **settings) -> dict:
            resp = self.put(url, **settings)
            return resp.json()

        def patch_json(self, url: str, **settings) -> dict:
            resp = self.patch(url, **settings)
            return resp.json()

    return ret_obj(**kwargs)



def bs4(text, url, obj=None, **kwargs):
    func = None
    if obj is None:
        from bs4 import BeautifulSoup

        func = BeautifulSoup
    else:
        func = obj

    k = {"markup": text, "features": "html.parser"}
    k.update(kwargs)

    return func(**k)


def html5_parser(text, url, obj=None, **kwargs):
    func = None
    if obj is None:
        from html5_parser import parse

        func = parse
    else:
        func = obj

    return func(text, **kwargs)


def modest(text, url, obj=None, **kwargs):
    func = None
    if obj is None:
        from selectolax.parser import HTMLParser

        func = HTMLParser
    else:
        func = obj

    return func(text, **kwargs)


def lexbor(text, url, obj=None, **kwargs):
    func = None
    if obj is None:
        from selectolax.parser import LexborHTMLParser

        func = LexborHTMLParser
    else:
        func = obj

    return func(text, **kwargs)


def lxml(text, url, obj=None, **kwargs):
    func = None
    if obj is None:
        from lxml.html import fromstring

        func = fromstring
    else:
        func = obj

    k = {"base_url": url}
    k.update(kwargs)

    return func(text, **k)


def reliq(text, url, obj=None, **kwargs):
    func = None
    if obj is None:
        from reliq import reliq as reliq_tree

        func = reliq_tree
    else:
        func = obj

    k = {"html": text, "ref": url}
    k.update(kwargs)

    return func(**k)



from typing import Optional, Tuple
import argparse
import re
import ast


def conv_curl_header_to_requests(src):
    r = re.search(r"^\s*([A-Za-z0-9_-]+)\s*:(.*)$", src)
    if r is None:
        return None
    return {r[1]: r[2].strip()}


def conv_curl_cookie_to_requests(src):
    r = re.search(r"^\s*([A-Za-z0-9_-]+)\s*=(.*)$", src)
    if r is None:
        return None
    return {r[1]: r[2].strip()}


def valid_header(src):
    r = conv_curl_header_to_requests(src)
    if r is None:
        raise argparse.ArgumentTypeError('Invalid header "{}"'.format(src))
    return r


def valid_cookie(src):
    r = conv_curl_cookie_to_requests(src)
    if r is None:
        raise argparse.ArgumentTypeError('Invalid cookie "{}"'.format(src))
    return r


def valid_browser(browser):
    import browser_cookie3

    try:
        return getattr(browser_cookie3, browser)
    except AttributeError:
        raise argparse.ArgumentTypeError('no such browser "{}"'.format(browser))


def valid_time(time):
    time = time.strip()

    suffix = time[-1:]
    weight = 1
    if not suffix.isdigit():
        time = time[:-1]
        match suffix:
            case "s":
                weight = 1
            case "m":
                weight = 60
            case "h":
                weight = 3600
            case "d":
                weight = 24 * 3600
            case _:
                raise argparse.ArgumentTypeError(
                    'incorrect time format "{}"'.format(time)
                )
    try:
        num = float(time)
    except ValueError:
        raise argparse.ArgumentTypeError('incorrect time format "{}"'.format(time))

    if num < 0:
        raise argparse.ArgumentTypeError('incorrect time format "{}"'.format(time))

    return num * weight


def arg_name(name, rename: list[Tuple[str]]) -> Optional[Tuple[str, bool]]:
    if name is None:
        return
    for i in rename:
        l = len(i)
        if l < 1 or l > 2:
            assert 0

        if i[0] != name:
            continue

        if l == 1:
            return
        else:
            return i[1], True
    return name, False


def args_longarg(arg: str, prefix: str, rename: list[Tuple[str]]):
    if (r := arg_name(arg, rename)) is None:
        return
    longarg = "--"
    if r[1]:
        return longarg + r[0]
    if len(prefix) > 0:
        longarg += prefix + "-"
    longarg += r[0]
    return longarg


def args_shortarg(
    shortarg: Optional[str], noshortargs: Optional[list[str]], rename: list[Tuple[str]]
) -> Optional[str]:
    if noshortargs:
        return
    if (r := arg_name(shortarg, rename)) is None:
        return
    return "-" + r[0]


def rename_normalize(rename):
    for i, j in enumerate(rename):
        if isinstance(j, str):
            val = j.lstrip("-")
            rename[i] = (val,)
        else:
            rename[i] = tuple([k.lstrip("-") for k in j])


def args_section(
    parser,
    name: str = "Request settings",
    noshortargs: bool = False,
    rename: list[Tuple[str, str] | Tuple[str] | str] = [],
    prefix: str = "",
):
    section = parser.add_argument_group(name)

    rename_normalize(rename)

    def add(shortarg, longarg, help, **kwargs):
        shortarg = args_shortarg(shortarg, noshortargs, rename)
        longarg = args_longarg(longarg, prefix, rename)

        if shortarg is None and longarg is None:
            return

        r = shortarg if shortarg is not None else longarg

        help = help.replace("{.}", r)

        a = [i for i in (shortarg, longarg) if i is not None]
        section.add_argument(*a, default=None, help=help, **kwargs)

    add(
        "w",
        "wait",
        "Sets waiting time for each request",
        metavar="TIME",
        type=valid_time,
    )
    add(
        "W",
        "wait-random",
        "Sets random waiting time for each request to be at max MILISECONDS",
        metavar="MILISECONDS",
        type=int,
    )
    add(
        "r",
        "retries",
        "Sets number of retries for failed request to NUM",
        metavar="NUM",
        type=int,
    )
    add(
        None,
        "retry-wait",
        "Sets interval between each retry",
        metavar="TIME",
        type=valid_time,
    )
    add(
        None,
        "force-retry",
        "Retry no matter the error",
        action="store_true",
    )
    add(
        "m",
        "timeout",
        "Sets request timeout",
        metavar="TIME",
        type=valid_time,
    )
    add(
        "k",
        "insecure",
        "Ignore ssl errors",
        action="store_false",
    )
    add(
        "L",
        "location",
        "Allow for redirections, can be dangerous if credentials are passed in headers",
        action="store_true",
    )
    add(
        "A",
        "user-agent",
        "Sets custom user agent",
        metavar="UA",
        type=str,
    )
    add(
        "x",
        "proxies",
        'Set requests proxies dictionary, e.g. {.} \'{"http":"127.0.0.1:8080","ftp":"0.0.0.0"}\'',
        metavar="DICT",
        type=lambda x: dict(ast.literal_eval(x)),
    )
    add(
        "H",
        "header",
        "Set curl style header, can be used multiple times e.g. {.} 'User: Admin' {.} 'Pass: 12345'",
        metavar="HEADER",
        type=valid_header,
        action="append",
    )
    add(
        "b",
        "cookie",
        "Set curl style cookie, can be used multiple times e.g. {.} 'auth=8f82ab' {.} 'PHPSESSID=qw3r8an829'",
        metavar="COOKIE",
        type=valid_cookie,
        action="append",
    )
    add(
        "B",
        "browser",
        "Get cookies from specified browser e.g. {.} firefox",
        metavar="BROWSER",
        type=valid_browser,
    )

    return section


def finish_cookies(cookies):
    ret = {}
    if cookies is None:
        return ret
    for i in cookies:
        ret.update(i)
    return ret


def finish_headers(headers, cookies):
    ret = {}
    if headers is None:
        return ret

    for i in headers:
        ret.update(i)

    cookie = headers.get("Cookie")
    if cookie is None:
        return ret

    ret.pop("Cookie")
    for i in cookie.split(";"):
        pair = i.split("=")
        name = pair[0].strip()
        val = None
        if len(pair) > 1:
            val = pair[1].strip()
        cookies.update({name: val})
    return ret


def args_session(
    session,
    args,
    prefix: str = "",
    rename: list[Tuple[str, str] | Tuple[str] | str] = [],
    **settings,
):
    prefix = prefix.replace("-", "_")
    rename_normalize(rename)

    def argval(name: str):
        if (r := arg_name(name, rename)) is None:
            return
        name = r[0]
        if not r[1] and len(prefix) > 0 and len(name) > 1:
            return getattr(args, prefix + "_" + name)
        else:
            return getattr(args, name)

    cookies = finish_cookies(argval("cookie"))
    headers = finish_headers(argval("header"), cookies)

    settings["headers"] = headers
    settings["cookies"] = cookies

    def setarg(longarg: str, shortarg: str = None, dest: str = None):
        name = longarg
        if (value := argval(name)) is None:
            if (value := argval(shortarg)) is None:
                return

        if dest is not None:
            name = dest
        settings[name] = value

    setarg("timeout")
    setarg("proxies")
    setarg("insecure", dest="verify")
    setarg("location", dest="allow_redirects")
    setarg("retries")
    setarg("retry_wait")
    setarg("force_retry")
    setarg("wait")
    setarg("wait_random")
    setarg("user_agent")
    setarg("browser")

    session.set_settings(settings, remove=False)



import random

useragents = {
    "desktop": {
        "windows": {
            "firefox": [
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:129.0) Gecko/20100101 Firefox/129.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:130.0) Gecko/20100101 Firefox/130.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:131.0) Gecko/20100101 Firefox/131.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:132.0) Gecko/20100101 Firefox/132.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:133.0) Gecko/20100101 Firefox/133.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:134.0) Gecko/20100101 Firefox/134.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:135.0) Gecko/20100101 Firefox/135.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:136.0) Gecko/20100101 Firefox/136.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:137.0) Gecko/20100101 Firefox/137.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:137.0) Gecko/20100101 Firefox/137.9",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:138.0) Gecko/20100101 Firefox/138.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:139.0) Gecko/20100101 Firefox/139.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:140.0) Gecko/20100101 Firefox/140.0",
            ],
            "chrome": [
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.6788.71 Safari/537.36",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36 Config/92.2.2671.19",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36 Trailer/90.3.9982.83",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6805.69 Safari/537.36",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.6784.68 Safari/537.36",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.6831.68 Safari/537.36",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.6833.84 Safari/537.36",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6828.70 Safari/537.36",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.7049.43 Safari/537.36",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36 Trailer/93.3.3570.29",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.7103.48 Safari/537.36",
            ],
            "opera": [
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.0.0 Safari/537.36 OPR/118.0.0.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36 OPR/120.0.0.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36 OPR/119.0.0.0",
            ],
            "edge": [
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36 Edg/131.0.0.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36 Edg/132.0.0.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6795.81 Safari/537.36 Edg/134.0.2779.105",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6815.78 Safari/537.36 Edg/133.0.2795.110",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.0.0 Safari/537.36 Edg/133.0.0.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.6786.62 Safari/537.36 Edg/134.0.2875.99",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36 Edg/134.0.0.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6793.79 Safari/537.36 Edg/132.0.2803.102",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6826.67 Safari/537.36 Edg/134.0.2784.107",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6832.82 Safari/537.36 Edg/134.0.2842.106",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36 Edg/136.0.0.0",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36 Edg/136.0.0.0 Viewer/99.9.8447.83",
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36 Edg/137.0.0.0",
                "Mozilla/5.0 (Windows NT 10.0; WOW64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6809.77 Safari/537.36 Edg/132.0.2739.110",
                "Mozilla/5.0 (Windows NT 10.0; WOW64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.6788.67 Safari/537.36 Edg/133.0.2795.103",
                "Mozilla/5.0 (Windows NT 10.0; WOW64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6782.80 Safari/537.36 Edg/132.0.2841.105",
            ],
        },
        "linux": {
            "firefox": [
                "Mozilla/5.0 (X11; Linux x86_64; rv:131.0) Gecko/20100101 Firefox/131.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:132.0) Gecko/20100101 Firefox/132.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:133.0) Gecko/20100101 Firefox/133.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:134.0) Gecko/20100101 Firefox/134.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:135.0) Gecko/20100101 Firefox/135.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:136.0) Gecko/20100101 Firefox/136.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:137.0) Gecko/20100101 Firefox/137.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:138.0) Gecko/20100101 Firefox/138.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:139.0) Gecko/20100101 Firefox/139.0",
                "Mozilla/5.0 (X11; Linux x86_64; rv:140.0) Gecko/20100101 Firefox/140.0",
            ],
            "chrome": [
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6665.58 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6807.78 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6807.85 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6830.67 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6785.82 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6820.66 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6822.76 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.0.0 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.6833.59 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6812.70 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6815.83 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Iron Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Safari/537.36",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/138.0.0.0 Safari/537.36",
            ],
            "opera": [
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Safari/537.36 OPR/113.0.0.0",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36 OPR/114.0.0.0",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Safari/537.36 OPR/115.0.0.0",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36 OPR/117.0.0.0",
                "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.0.0 Safari/537.36 OPR/118.0.0.0",
            ],
        },
        "macos": {
            "firefox": [
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:130.0) Gecko/20100101 Firefox/130.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:131.0) Gecko/20100101 Firefox/131.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:132.0) Gecko/20100101 Firefox/132.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:133.0) Gecko/20100101 Firefox/133.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:134.0) Gecko/20100101 Firefox/134.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:135.0) Gecko/20100101 Firefox/135.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:136.0) Gecko/20100101 Firefox/136.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:137.0) Gecko/20100101 Firefox/137.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:138.0) Gecko/20100101 Firefox/138.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:139.0) Gecko/20100101 Firefox/139.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:140.0) Gecko/20100101 Firefox/140.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 13.6; rv:132.0) Gecko/20100101 Firefox/132.0",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14.1; rv:131.0) Gecko/20010101 Firefox/131.0",
            ],
            "chrome": [
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.6291.223 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_11) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.6806.80 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6803.79 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 12.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6823.77 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 12_2_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6798.84 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6800.71 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.6789.64 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14.3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6804.78 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14.4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6804.85 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6832.61 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6820.63 Safari/537.36",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 14) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/133.0.6796.65 Safari/537.36",
            ],
            "safari": [
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.4.1 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.4 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.5 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.6 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.8.1 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.8 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.0.1 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.0 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.1.1 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.1 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.2 Safari/605.1.15",
                "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.3.1 Safari/605.1.15",
            ],
        },
    },
    "phone": {
        "android": {
            "chrome": [
                "Mozilla/5.0 (Linux; Android 10.2; SM-T875) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6808.57 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/115.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/135.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/136.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; K) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/137.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; MI 8 Lite) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; Redmi Note 3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/116.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 10; V1921A; Build/QP1A.190711.020) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/124.0.4280.141 Mobile Safari/537.36 Firefox-KiToBrowser/124.0",
                "Mozilla/5.0 (Linux; Android 11; CPH2099) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 11; M2006C3LG) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 11; M2006C3LG) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Mobile Safari/537.36 uacq",
                "Mozilla/5.0 (Linux; Android 12; Lenovo TB-X606F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6785.62 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 12; M40_EEA) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6820.57 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 12; Redmi Note 9 Pro) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 12; SM-T865) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6802.83 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 13; 23054RA19C) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 13; A850) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6786.73 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 13) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Mobile Safari/537.36 uacq",
                "Mozilla/5.0 (Linux; Android 13; Lenovo TB-X505L) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.6788.74 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 13; M2012K11AG) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 13; SM-A037U) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/112.0.0.0 Mobile Safari/537.36 uacq",
                "Mozilla/5.0 (Linux; Android 13; SM-G973F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.6820.72 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 14; moto g54 5G) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/110.0.0.0 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 14; TX6s) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.6308.220 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 14; Z832 Build/MMB29M) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.6613.127 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 4.0.4; Galaxy Nexus Build/IMM76B) AppleWebKit/537.36 (KHTML, like Gecko; Mediapartners-Google) Chrome/131.0.6778.139 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 4.0.4; Galaxy Nexus Build/IMM76B) AppleWebKit/537.36 (KHTML, like Gecko; Mediapartners-Google) Chrome/134.0.6998.165 Mobile Safari/537.36",
                "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/127.0.6533.99 Mobile Safari/537.36 (compatible; Google-Safety; +http://www.google.com/bot.html)",
                "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 5X Build/MMB29P) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/131.0.6778.139 Mobile Safari/537.36 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
                "Mozilla/5.0 (Linux; Android 9; ASUS_X00TD; Flow) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/359.0.0.288 Mobile Safari/537.36",
            ],
            "firefox": [
                "Mozilla/5.0 (Android 10; Mobile; rv:128.0) Gecko/128.0 Firefox/128.0",
                "Mozilla/5.0 (Android 10; Mobile; rv:129.0) Gecko/129.0 Firefox/129.0",
                "Mozilla/5.0 (Android 10; Mobile; rv:131.0) Gecko/131.0 Firefox/131.0",
                "Mozilla/5.0 (Android 10; Mobile; rv:133.0) Gecko/133.0 Firefox/133.0",
                "Mozilla/5.0 (Android 10; Mobile; rv:134.0) Gecko/134.0 Firefox/134.0",
                "Mozilla/5.0 (Android 10; Mobile; rv:137.0) Gecko/137.0 Firefox/137.0",
                "Mozilla/5.0 (Android 11; Mobile; rv:128.0) Gecko/128.0 Firefox/128.0",
                "Mozilla/5.0 (Android 11; Mobile; rv:129.0) Gecko/129.0 Firefox/129.0",
                "Mozilla/5.0 (Android 11; Mobile; rv:134.0) Gecko/134.0 Firefox/134.0",
                "Mozilla/5.0 (Android 11; Mobile; rv:136.0) Gecko/136.0 Firefox/136.0",
                "Mozilla/5.0 (Android 12; Mobile; rv:129.0) Gecko/129.0 Firefox/129.0",
                "Mozilla/5.0 (Android 12; Mobile; rv:133.0) Gecko/133.0 Firefox/133.0",
                "Mozilla/5.0 (Android 12; Mobile; rv:135.0) Gecko/135.0 Firefox/135.0",
                "Mozilla/5.0 (Android 12; Mobile; rv:137.0) Gecko/137.0 Firefox/137.0",
                "Mozilla/5.0 (Android 13; Mobile; rv:128.0) Gecko/128.0 Firefox/128.0",
                "Mozilla/5.0 (Android 13; Mobile; rv:129.0) Gecko/129.0 Firefox/129.0",
                "Mozilla/5.0 (Android 13; Mobile; rv:130.0) Gecko/130.0 Firefox/130.0",
                "Mozilla/5.0 (Android 13; Mobile; rv:133.0) Gecko/133.0 Firefox/133.0",
                "Mozilla/5.0 (Android 13; Mobile; rv:134.0) Gecko/134.0 Firefox/134.0",
                "Mozilla/5.0 (Android 13; Mobile; rv:138.0) Gecko/138.0 Firefox/138.0",
                "Mozilla/5.0 (Android 14; Mobile; rv:128.0) Gecko/128.0 Firefox/128.0",
                "Mozilla/5.0 (Android 14; Mobile; rv:129.0) Gecko/129.0 Firefox/129.0",
                "Mozilla/5.0 (Android 14; Mobile; rv:131.0) Gecko/131.0 Firefox/131.0",
                "Mozilla/5.0 (Android 14; Mobile; rv:133.0) Gecko/133.0 Firefox/133.0",
                "Mozilla/5.0 (Android 14; Mobile; rv:137.0) Gecko/137.0 Firefox/137.0",
                "Mozilla/5.0 (Android 14; Mobile; rv:138.0) Gecko/138.0 Firefox/138.0",
                "Mozilla/5.0 (Android 15; Mobile; rv:135.0) Gecko/135.0 Firefox/135.0",
            ],
        },
        "ios": {
            "safari": [
                "Mozilla/5.0 (iPhone; CPU iPhone OS 14_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.4 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 15_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.4 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 15_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/15.6 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 16_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.4.26 Mobile/Y650R1 Safari/619.10",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 16_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.5 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 16_6_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 16_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 16_7_2 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.6 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 16 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.5.17 Mobile/N1WEFQ Safari/617.32.2",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_4_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.4.1 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_5_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.5 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_5_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/16.1.15 Mobile/GCUL1T Safari/618.5.14",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.5 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_6_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.6.1 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_6_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.6 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 18_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.0 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 18_1_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.1.1 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 18_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.1 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 18_3_2 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.3.1 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 18_4_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/18.4 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU OS 17_5_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/17.4.1 Mobile/14E304 Safari/605.1.15",
            ],
            "firefox": [
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_6_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) FxiOS/129.2 Mobile/15E148 Safari/605.1.15"
            ],
            "chrome": [
                "Mozilla/5.0 (iPhone; CPU iPhone OS 16_2 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/125.0.6422.51 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/124.0.6367.111 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/125.0.6422.51 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_5_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/125.0.6422.51 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/126.0.6478.54 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_5 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/128.0.6613.98 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/127.0.6533.77 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 17_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/128.0.6613.98 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 18_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/129.0.6668.21 Mobile/15E148 Safari/604.1",
                "Mozilla/5.0 (iPhone; CPU iPhone OS 18_1_0 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) CriOS/131.0.6778.154 Mobile/15E148 Safari/604.1",
            ],
        },
    },
    "bot": {
        "google": [
            "Mozilla/5.0 AppleWebKit/537.36 (KHTML, like Gecko; compatible; Googlebot/2.1; +http://www.google.com/bot.html) Chrome/124.0.6367.201 Safari/537.36",
            "Mozilla/5.0 AppleWebKit/537.36 (KHTML, like Gecko; compatible; Googlebot/2.1; +http://www.google.com/bot.html) Chrome/134.0.6998.165 Safari/537.36",
            "Mozilla/5.0 AppleWebKit/537.36 (KHTML, like Gecko; compatible; Googlebot/2.1; +http://www.google.com/bot.html) Safari/537.36",
            "Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)",
            "Googlebot/2.1 (+http://www.google.com/bot.html)",
        ],
        "bing": [
            "Mozilla/5.0 AppleWebKit/537.36 (KHTML, like Gecko; compatible; bingbot/2.0; +http://www.bing.com/bingbot.htm) Chrome/116.0.1938.76 Safari/537.36",
            "Mozilla/5.0 (compatible; bingbot/2.0; +http://www.bing.com/bingbot.htm)",
        ],
        "yandex": [
            "Mozilla/5.0 (compatible; YandexAccessibilityBot/3.0; +http://yandex.com/bots)",
            "Mozilla/5.0 (compatible; YandexBot/3.0; +http://yandex.com/bots)",
            "Mozilla/5.0 (compatible; YandexMedia/3.0; +http://yandex.com/bots)",
        ],
        "duckduckgo": [
            "DuckDuckBot/1.1; (+http://duckduckgo.com/duckduckbot.html)",
        ],
    },
}


def newagent_edges(ls):
    if isinstance(ls, list):
        return ls
    else:
        ret = []
        for i in ls.keys():
            ret += newagent_edges(ls[i])
        return ret


def newagent_deep(ls, paths):
    if len(paths) == 0:
        return newagent_edges(ls)

    x = paths[0]
    after = paths[1:]

    if isinstance(x, str):
        return newagent_deep(ls[x], after)
    else:
        ret = []
        for i in x:
            if isinstance(i, str):
                ret += newagent_deep(ls[i], after)
            else:
                ret += newagent_deep(ls, i)
        return ret


def newagent(*paths) -> str:
    if len(paths) == 1 and isinstance(paths[0], str):
        return paths[0]

    if len(paths) == 0:
        acc = newagent_edges(useragents)
    else:
        acc = []
        for i in paths:
            if isinstance(i, str):
                acc += newagent_edges(useragents[i])
            else:
                acc += newagent_deep(useragents, i)

    return random.choice(acc)



from typing import Optional, Callable
from pathlib import Path
import io


class Logger:
    def ispath(self):
        return isinstance(self.source, str) or isinstance(self.source, Path)

    def __init__(self, dest: list | str | Path | io.TextIOWrapper | Callable):
        self.source = dest
        self.dest = dest
        if self.ispath():
            self.dest = open(dest, "w")

    def log(self, *args):
        dest = self.dest
        if isinstance(dest, Callable):
            dest(args)
            return

        if isinstance(dest, list):
            dest.append(args)
            return

        s = "\t".join(str(i) for i in args)

        if isinstance(dest, io.TextIOWrapper):
            dest.write(s)
            dest.write("\n")
            return

        raise Exception("Unknown type")

    def __del__(self):
        if self.ispath():
            self.dest.close()


def create_logger(
    dest: Optional[list | str | Path | io.TextIOWrapper | Callable],
) -> Optional[Callable]:
    if dest is None:
        return

    log = Logger(dest)
    return log.log


def simple_logger(dest: list | str | Path | io.TextIOWrapper | Callable):
    class Lg(Logger):
        def log(self, *args):
            if len(args) != 3:
                assert 0

            super().log(args[1])

    log = Lg(dest)
    return lambda x: log.log(*x)
